/*
  Rainbow Serpent, a UCI chess playing engine based on Stockfish 
  which in turn is derived from Glaurung 2.1
  Copyright (C) 2004-2008 Tord Romstad (Glaurung author)
  Copyright (C) 2008-2012 Marco Costalba, Joona Kiiski, Tord Romstad
  Copyright source changes from Stockfish (C) 2011-2012 Eelco de Groot

  Rainbow Serpent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Rainbow Serpent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <cassert>

#include "bitboard.h"
#include "bitcount.h"
#include "pawns.h"
#include "position.h"

namespace {

  #define V Value
  #define S(mg, eg) make_score(mg, eg)

  // Doubled pawn penalty by opposedOnAdjacentFiles flag and file
  const Score DoubledPawnPenalty[2][8] = {
  { S(13, 43), S(20, 48), S(23, 48), S(23, 48),
    S(23, 48), S(23, 48), S(20, 48), S(13, 43) },
  { S(10, 40), S(15, 40), S(20, 40), S(20, 40),
    S(20, 40), S(20, 40), S(20, 40), S(10, 40) }};

  // Isolated pawn penalty by opposed flag and file
  const Score IsolatedPawnPenalty[2][8] = {
  { S(37, 45), S(54, 52), S(60, 52), S(60, 52),
    S(60, 52), S(60, 52), S(54, 52), S(37, 45) },
  { S(25, 30), S(36, 35), S(40, 35), S(40, 35),
    S(40, 35), S(40, 35), S(36, 35), S(25, 30) }};
  
  // Half isolated pawn penalty by opposed flag and file.
  // Constants for A- and H-file not actually used because
  // those pawns are scored as isolated pawns.

  const Score HalfIsolatedPawnPenalty[2][8] = {
  { S(27, 35), S(44, 42), S(50, 42), S(50, 42),
    S(50, 42), S(50, 42), S(44, 42), S(27, 35) },
  { S(15, 20), S(26, 25), S(30, 25), S(30, 25),
    S(30, 25), S(30, 25), S(26, 25), S(15, 20) }};

  // Backward pawn penalty by opposed flag and file
  const Score BackwardPawnPenalty[2][8] = {
  { S(30, 42), S(43, 46), S(49, 46), S(49, 46),
    S(49, 46), S(49, 46), S(43, 46), S(30, 42) },
  { S(20, 28), S(29, 31), S(33, 31), S(33, 31),
    S(33, 31), S(33, 31), S(29, 31), S(20, 28) }};

  // Pawn chain membership bonus by file
  const Score ChainBonus[8] = {
    S(17, 2), S(19, 2), S(19, 2), S(20, 2),
    S(20, 2), S(19, 2), S(19, 2), S(17, 2)
  };

  // Candidate passed pawn bonus by rank
  const Score CandidateBonus[8] = {
    S( 0, 0), S( 6, 13), S(6,13), S(14,29),
    S(34,68), S(83,166), S(0, 0), S( 0, 0)
  };

  const Score PawnStructureWeight = S(233, 201);

  // Weakness of our pawn shelter in front of the king indexed by [king pawn][rank]
  const Value ShelterWeakness[2][8] =
  { { V(141), V(0), V(38), V(102), V(128), V(141), V(141) },
    { V( 61), V(0), V(16), V( 44), V( 56), V( 61), V( 61) } };

  // Danger of enemy pawns moving toward our king indexed by [pawn blocked][rank]
  const Value StormDanger[2][8] =
  { { V(26), V(0), V(128), V(51), V(26) },
    { V(13), V(0), V( 64), V(25), V(13) } };

  // Max bonus for king safety. Corresponds to start position with all the pawns
  // in front of the king and no enemy pawn on the horizont.
  const Value MaxSafetyBonus = V(263);

  #undef S
  #undef V  
}


/// PawnTable::probe() takes a position object as input, computes a PawnEntry 
/// object, and returns a pointer to it. The result is also stored in a hash
/// table, so we don't have to recompute everything when the same pawn structure
/// occurs again.

PawnEntry* PawnTable::probe(const Position& pos) {

  Key key = pos.pawn_key();
  PawnEntry* e = entries[key];

  // If e->key matches the position's pawn hash key, it means that we
  // have analysed this pawn structure before, and we can simply return
  // the information we found the last time instead of recomputing it.
  if (e->key == key)
      return e;

  // Initialize PawnEntry
  e->key = key;
  e->passedPawns[WHITE] = e->passedPawns[BLACK] = 0;
  e->chainbasePawns[WHITE] = e->chainbasePawns[BLACK] = 0;
  e->kingSquares[WHITE] = e->kingSquares[BLACK] = SQ_NONE;
  e->halfOpenFiles[WHITE] = e->halfOpenFiles[BLACK] = 0xFF;

  // Calculate pawn attacks
  Bitboard wPawns = pos.pieces(WHITE, PAWN);
  Bitboard bPawns = pos.pieces(BLACK, PAWN);
  e->pawnAttacks[WHITE] = ((wPawns & ~FileHBB) << 9) | ((wPawns & ~FileABB) << 7);
  e->pawnAttacks[BLACK] = ((bPawns & ~FileHBB) >> 7) | ((bPawns & ~FileABB) >> 9);

  // Evaluate pawns for both colors and weight the result
  e->value =  evaluate_pawns<WHITE>(pos, wPawns, bPawns, e)
            - evaluate_pawns<BLACK>(pos, bPawns, wPawns, e);

  e->value = apply_weight(e->value, PawnStructureWeight);

  return e;
}


/// PawnTable::evaluate_pawns() evaluates each pawn of the given color

template<Color Us>
Score PawnTable::evaluate_pawns(const Position& pos, Bitboard ourPawns,
                                    Bitboard theirPawns, PawnEntry* e) {

  const Color Them = (Us == WHITE ? BLACK : WHITE);

  Bitboard b, bb;
  Square s, blockSq;
  File f;
  Rank r;
  bool passed, isolated, leftIsolated, rightIsolated, doubled, opposed, opposedOnAdjacentFiles;
  bool chain, backward, candidate, blocked, indirectlyBlocked, weak;
  bool attackedBlocksquare, attacked, attackedGote, attackedSente, defended, chainBase;
  int candidatesupportcount, candidaterank;
  Score value = SCORE_ZERO, pawnValue, backwardPenalty = SCORE_ZERO;
  const Square* pl = pos.piece_list(Us, PAWN);

  // Loop through all pawns of the current color and score each pawn
  while ((s = *pl++) != SQ_NONE)
  {
      assert(pos.piece_on(s) == make_piece(Us, PAWN));
	  
	  blockSq = s + pawn_push(Us);
      f = file_of(s);
      r = rank_of(s);

      // This file cannot be half open
      e->halfOpenFiles[Us] &= ~(1 << f);

      // Our rank plus previous one. Used for chain detection
      b = rank_bb(r) | rank_bb(Us == WHITE ? r - Rank(1) : r + Rank(1));
	  
	  // Pawn blocked by enemy pawn?
	  blocked = theirPawns & blockSq;
	  indirectlyBlocked = blocked;

	  // Square in front attacked by enemy pawn?
	  attackedBlocksquare = pos.attacks_from<PAWN>(blockSq, Us) & theirPawns;
	  
	  // Pawn attacked by enemy pawn?
	  attacked = pos.attacks_from<PAWN>(s, Us) & theirPawns;
	  attackedGote = attacked && Them == pos.side_to_move();
	  attackedSente = attacked && !attackedGote;

      // Flag the pawn as passed, isolated, doubled or member of a pawn
      // chain (but not the backward one).
      doubled  =  ourPawns   & forward_bb(Us, s);
      opposed  =  theirPawns & forward_bb(Us, s);
	  passed   = !opposed & !(theirPawns & passed_pawn_mask(Us, s));
      isolated = !(ourPawns & adjacent_files_bb(f));
	  leftIsolated   = !isolated && !passed && !((f == FILE_A) | (f == FILE_H)) && !(ourPawns  & left_neighboring_files_bb(f));
	  rightIsolated  = !isolated && !passed && !leftIsolated && !((f == FILE_A) | (f == FILE_H)) && !(ourPawns  & right_neighboring_files_bb(f));
      chain      =  ourPawns   & adjacent_files_bb(f) & b;
	  defended   =  chain && (ourPawns & adjacent_files_bb(f) & rank_bb(Us == WHITE ? r - Rank(1) : r + Rank(1)));
	  weak       = !(chain && (ourPawns & adjacent_files_bb(f) & rank_bb(r))); // blockSq is undefended
	  chainBase  = (isolated && blocked) || (!defended && weak && (ourPawns & adjacent_files_bb(f) & (rank_bb(Us == WHITE ? (r + Rank(1)) : (r - Rank(1))) | rank_bb(Us == WHITE ? (r + Rank(2)) : (r - Rank(2))))));
	  opposedOnAdjacentFiles = theirPawns & passed_pawn_mask(Us, s);

	  
	  // Test for candidate passed pawn
	  candidate = false;
	  candidatesupportcount = 0;
	  bb = attack_span_mask(Them, blockSq) & ourPawns;
	  if (!blocked && !passed && !opposed && bb)
	  {
		  candidatesupportcount = popcount<Max15>(bb) - popcount<Max15>(attack_span_mask(Us, s) & theirPawns);
		  candidate = candidatesupportcount >= 0;
	  }

      // Test for backward pawn
      backward = false;

      // If the pawn is a candidate pawn, passed, isolated, or member of a pawn chain
      // it cannot be backward. If there are friendly pawns beside or behind on neighboring
      // files it cannot be backward either.
      if (!(candidate | passed | isolated | chain | bb))
          //&& !(ourPawns & attack_span_mask(opposite_color(Us), s))
          //&& !(pos.attacks_from<PAWN>(s, Us) & theirPawns))
      {
          // We now know that there are no friendly pawns beside or behind this
          // pawn on neighboring files. We now check whether the pawn is
          // backward by looking in the forward direction on the neighboring
          // files, and seeing whether we meet a friendly or an enemy pawn first.
          b = pos.attacks_from<PAWN>(s, Us);
		  backward = true;
		  
		  if (!blocked && (b & ourPawns))
		  {
			  if (Us == WHITE)
			  {
				  if ((relative_rank(Us, s) >= RANK_4) || !((b | (b << 8)) & theirPawns)) backward = false;
			  }
			  else
			  {
				  if ((relative_rank(Us, s) >= RANK_4) || !((b | (b >> 8)) & theirPawns)) backward = false;
			  }
			  /*
			  if (   backward
				  && opposed
				  && (count_1s<Max15>(b & ourPawns) == 2)
				  && (count_1s<Max15>((Us == WHITE ? (b << 8) : (b >> 8)) & theirPawns) == 2))
				  indirectlyBlocked = true; 
			  // This is maybe not watertight but just an indicator that
			  // a breakthrough will be difficult. Expensive calculation
			  */
		  }
		  /*

          // Note that we are sure to find something because pawn is not passed
          // nor isolated, so loop is potentially infinite, but it isn't.
          while (!(b & (ourPawns | theirPawns)))
              Us == WHITE ? b <<= 8 : b >>= 8;

          // The friendly pawn needs to be at least two ranks closer than the enemy
          // pawn in order to help the potentially backward pawn advance.
          backward = (b | (Us == WHITE ? b << 8 : b >> 8)) & theirPawns;
		  
		  */
      }

      assert(opposed | passed | (attack_span_mask(Us, s) & theirPawns));
	  
	  pawnValue = SCORE_ZERO;

      // Mark the pawn as passed. Pawn will be further scored in evaluation
      // because we need full attack info to evaluate passed pawns. Only the
      // frontmost passed pawn on each file is considered a true passed pawn.
      if (passed && !doubled)
		  e->passedPawns[Us] |= s;
	  // isolated passed pawns are scored dynamically only!
	  
	  // Mark and score chainbase pawns
	  if (chainBase)
	  {
		  e->chainbasePawns[Us] |= s;
		  if (attackedSente)
			  pawnValue -= make_score(27, 39);
		  if (attackedGote)
			  pawnValue -= make_score(30, 45);
		  if (!attackedSente && !attackedGote  && !blocked)
			  pawnValue += make_score(20, 20 - square_distance(pos.king_square(Us), s));
	  }

      // Score this pawn
      if (isolated && !passed)
	  {
          pawnValue -= IsolatedPawnPenalty[opposed][f];
		  if (doubled)  //[was: && !opposedOnAdjacentFiles), && !(!opposed && (theirPawns & adjacent_files_bb(f))))]
			  pawnValue -= IsolatedPawnPenalty[opposed][f];
		  if (blocked)
			  pawnValue -= make_score(0, 20 + 3 * square_distance(pos.king_square(Us), s)
											- 2 * square_distance(pos.king_square(Them), s)
											+ (opposedOnAdjacentFiles && !doubled ? 40 : 0)); //[Too complicated? FIXME]
	  }
	  
	  if (leftIsolated | rightIsolated) // passed pawns are already excluded
	  {
          pawnValue -= HalfIsolatedPawnPenalty[opposed][f];
		  if (doubled)
			  pawnValue -= HalfIsolatedPawnPenalty[opposed][f]/2;
		  if (backward)
			  pawnValue -= blocked ? HalfIsolatedPawnPenalty[true][f] : HalfIsolatedPawnPenalty[opposed][f]/2;
	  }

      if (doubled)
          pawnValue -= DoubledPawnPenalty[opposedOnAdjacentFiles][f];

      if (backward)
	  {
          backwardPenalty = BackwardPawnPenalty[opposed][f];
		  if (blocked)
			  backwardPenalty += make_score(relative_rank(Them, blockSq) * 2, relative_rank(Them, blockSq)); //[was -=]
		  if (attackedSente)
			  backwardPenalty -= BackwardPawnPenalty[opposed][f]/2;
		  if (attackedGote)
			  backwardPenalty += BackwardPawnPenalty[opposed][f]/2; // Pawn is not defended and capture will likely create a passed pawn
		  pawnValue -= backwardPenalty;
	  }

      if (chain)
          pawnValue += ChainBonus[f];
	  
	  if (weak | !defended)
	  {
		  if (!chain && !defended && (ourPawns & adjacent_files_bb(f) & rank_bb(Us == BLACK ? r - Rank(1) : r + Rank(1)))) // Not defended and defending
			  pawnValue -= attacked ? (Them == pos.side_to_move() ? 2*ChainBonus[f] : ChainBonus[f]) : ChainBonus[f]/2; // Too complicated FIXME
		  pawnValue -= (blocked || attackedBlocksquare) ? make_score(15, 20) : make_score(0, 0);
	  }

      if (candidate)
	  {
		  pawnValue += CandidateBonus[relative_rank(Us, s)];		  

		  if (candidatesupportcount > 0)
		  {
			  candidaterank = relative_rank(Us, s);
			  candidatesupportcount = candidatesupportcount * (doubled ? 0 : candidaterank) * (attacked ? (Them == pos.side_to_move() && !defended ? 0 : 5) : (!chain && attackedBlocksquare ? 10 : 20));
			  pawnValue += make_score(candidatesupportcount, 2 * candidatesupportcount - 3 * square_distance(pos.king_square(Us), s) + 2 * square_distance(pos.king_square(Them), s));			  
		  }		  
	  }

	  if (attackedSente && pawnValue < SCORE_ZERO)
		  pawnValue += defended ? make_score(15, 20) : make_score(20, 30);
	  if (attackedGote && pawnValue > SCORE_ZERO)
		  pawnValue -= defended ? pawnValue/3 : pawnValue/2;
	  if (attackedGote && pawnValue <= SCORE_ZERO)
		  pawnValue -= defended ? make_score(0, 0) : make_score(15, 20);
	  
	  value += pawnValue;
  }
  return value;
}

/// PawnEntry::shelter_storm() calculates shelter and storm penalties for the file
/// the king is on, as well as the two adjacent files.

template<Color Us>
Value PawnEntry::shelter_storm(const Position& pos, Square ksq) {

  const Color Them = (Us == WHITE ? BLACK : WHITE);

  Value safety = MaxSafetyBonus;
  Bitboard b = pos.pieces(PAWN) & (in_front_bb(Us, ksq) | rank_bb(ksq));
  Bitboard ourPawns = b & pos.pieces(Us) & ~rank_bb(ksq);
  Bitboard theirPawns = b & pos.pieces(Them);
  Rank rkUs, rkThem;
  File kf = file_of(ksq);

  kf = (kf == FILE_A) ? kf++ : (kf == FILE_H) ? kf-- : kf;

  for (int f = kf - 1; f <= kf + 1; f++)
  {
      // Shelter penalty is higher for the pawn in front of the king
      b = ourPawns & FileBB[f];
      rkUs = b ? rank_of(Us == WHITE ? lsb(b) : ~msb(b)) : RANK_1;
      safety -= ShelterWeakness[f != kf][rkUs];

      // Storm danger is smaller if enemy pawn is blocked
      b  = theirPawns & FileBB[f];
      rkThem = b ? rank_of(Us == WHITE ? lsb(b) : ~msb(b)) : RANK_1;
      safety -= StormDanger[rkThem == rkUs + 1][rkThem];
  }

  return safety;
}


/// PawnEntry::update_safety() calculates and caches a bonus for king safety. It is
/// called only when king square changes, about 20% of total king_safety() calls.

template<Color Us>
Score PawnEntry::update_safety(const Position& pos, Square ksq) {

  kingSquares[Us] = ksq;
  castleRights[Us] = pos.can_castle(Us);
  minKPdistance[Us] = 0;
  //int distancepenalty = 0;

  Bitboard pawns = pos.pieces(Us, PAWN);
  if (pawns)
      while (!(DistanceRingsBB[ksq][minKPdistance[Us]++] & pawns)) {}	  

  if (relative_rank(Us, ksq) > RANK_4)
      return kingSafety[Us] = make_score(0, -16 * minKPdistance[Us]);

  Value bonus = shelter_storm<Us>(pos, ksq);

  // If we can castle use the bonus after the castle if is bigger
  if (pos.can_castle(make_castle_right(Us, KING_SIDE)))
      bonus = std::max(bonus, shelter_storm<Us>(pos, relative_square(Us, SQ_G1)));

  if (pos.can_castle(make_castle_right(Us, QUEEN_SIDE)))
      bonus = std::max(bonus, shelter_storm<Us>(pos, relative_square(Us, SQ_C1)));
  
  else if (ksq == relative_square(Us, SQ_C1))
      bonus = std::min(bonus, shelter_storm<Us>(pos, relative_square(Us, SQ_B1)));

  return kingSafety[Us] = make_score(bonus, -16 * minKPdistance[Us]);
}

// Explicit template instantiation
template Score PawnEntry::update_safety<WHITE>(const Position& pos, Square ksq);
template Score PawnEntry::update_safety<BLACK>(const Position& pos, Square ksq);
